import importlib
import logging

from pyrsistent import field
from pyrsistent import m
from pyrsistent import PMap
from pyrsistent import pmap
from pyrsistent import PRecord

from task_processing.interfaces import TaskExecutor

log = logging.getLogger(__name__)


class Registry(PRecord):
    plugin_modules = field(type=PMap, initial=m(), factory=pmap)
    """
    A map of plugin names (str) to that plugins entry point (a module)

    We use this to allow the registry and plugins to dynamically
    change their behaviors. Currently just the plugins registering executors
    """

    def register_task_executor(self, name, task_executor_cls):
        """Helper method for adding an executor"""
        return self.transform(
            ('task_executors', name), lambda _: task_executor_cls
        )

    def register_deprecated_task_executor(self, name, task_executor_cls):
        """Helper method for adding an deprecated executor"""
        return self.transform(
            ('deprecated_task_executors', name), lambda _: task_executor_cls
        )

    def _executor_invariant(task_executors):
        """Invariant that task_executors must be TaskExecutors"""
        return (
            all(issubclass(v, TaskExecutor) for v in task_executors.values()),
            'task_executors must always be a TaskExecutor'
        )

    task_executors = field(
        type=PMap, initial=m(), factory=pmap,
        invariant=_executor_invariant,
    )
    deprecated_task_executors = field(
        type=PMap, initial=m(), factory=pmap,
        invariant=_executor_invariant,
    )
    """
    A map of task executor names (str) to a class definition which implements
    :class:`task_processing.interfaces.TaskExecutor`
    """


class TaskProcessor:
    """Main Entry Point for Task Processing plugins

    This is the entrypoint to registering plugins (which know how to
    actually run tasks) and creating instances of TaskExecutors. Note that
    since most customers of this library will be importing this class into
    their code, let's try to make it Python2/3 compatible.

    Typical use would be:

        tp = TaskProcessor()
        # registers 'myplugin'
        tp.load_plugin('my.plugin')
        # ... within the my.plugin.register_plugin method:
        registry.register_task_executor('foobar', FooBarExecutor)

        # register 'another'
        tp.load_plugin('another.plugin')

        # Later...
        executor = tp.executor_from_config(
            provider='foobar',
            provider_config={'specific': 'config', 'for', 'plugin'}
        )
        executor.run()
    """

    registry = Registry()

    def load_plugin(self, provider_module):
        module = importlib.import_module(provider_module)
        plugin_name = module.TASK_PROCESSING_PLUGIN

        name_update = Registry(plugin_modules=m().set(plugin_name, module))
        plugin_update = module.register_plugin(Registry())

        def conflict_check(old, new):
            # Any PMap in the registry ought have no duplicate keys,
            # if they do that means two plugins tried to define the same
            # executor or something
            if type(old) == PMap:
                conflicts = set(old.keys()) & set(new.keys())
                if conflicts:
                    raise ValueError(
                        '{0} is trying to register elements that already '
                        'exist in the registry: {1}'.format(
                            plugin_name, conflicts
                        )
                    )
            return old.update(new)

        self.registry = self.registry.update_with(
            conflict_check, plugin_update, name_update
        )

    def executor_cls(self, provider):
        """Called by users of TaskProcessing to obtain :class:`TaskExecutor`s

        :param str provider: The kind of executor the user wants to run.
        :param dict provider_config: The arguments needed to instantiate
            the provider.
        """
        if provider in self.registry.task_executors:
            return self.registry.task_executors[provider]
        elif provider in self.registry.deprecated_task_executors:
            log.warning(
                f'{provider} is a deprecated executor and will be removed in the future')
            return self.registry.deprecated_task_executors[provider]
        else:
            raise ValueError(
                '{0} provider not registered; available providers: {1}'.format(
                    provider, self.registry.task_executors.keys().tolist()
                )
            )

    def executor_from_config(self, provider, provider_config=None):
        if provider_config is None:
            provider_config = dict()

        return self.executor_cls(provider)(**provider_config)



from typing import List
from typing import Tuple

from task_processing.plugins.mesos.constraints import attributes_match_constraints
from task_processing.plugins.mesos.mesos_executor import MesosExecutor
from task_processing.plugins.mesos.mesos_executor import MesosExecutorCallbacks
from task_processing.plugins.mesos.resource_helpers import allocate_task_resources
from task_processing.plugins.mesos.resource_helpers import ResourceSet
from task_processing.plugins.mesos.resource_helpers import task_fits
from task_processing.plugins.mesos.task_config import MesosTaskConfig
from task_processing.plugins.mesos.translator import make_mesos_task_info
from task_processing.plugins.mesos.translator import mesos_update_to_event


def get_tasks_for_offer(
    task_configs: List[MesosTaskConfig],
    offer_resources: ResourceSet,
    offer_attributes: dict,
    role: str,
) -> Tuple[List[MesosTaskConfig], List[MesosTaskConfig]]:

    tasks_to_launch, tasks_to_defer = [], []

    for task_config in task_configs:
        if (task_fits(task_config, offer_resources) and
                attributes_match_constraints(offer_attributes, task_config.constraints)):
            prepared_task_config, offer_resources = allocate_task_resources(
                task_config,
                offer_resources,
            )
            tasks_to_launch.append(prepared_task_config)
        else:
            tasks_to_defer.append(task_config)

    return tasks_to_launch, tasks_to_defer


class MesosTaskExecutor(MesosExecutor):
    TASK_CONFIG_INTERFACE = MesosTaskConfig

    def __init__(self, role, *args, **kwargs) -> None:
        super().__init__(
            role,
            MesosExecutorCallbacks(
                get_tasks_for_offer,
                mesos_update_to_event,
                make_mesos_task_info,
            ),
            *args,
            **kwargs,
        )



import logging
import time
from queue import Queue

from task_processing.interfaces.runner import Runner

log = logging.getLogger(__name__)


class Sync(Runner):
    def __init__(self, executor):
        self.executor = executor
        self.TASK_CONFIG_INTERFACE = executor.TASK_CONFIG_INTERFACE
        self.queue = Queue()

    def kill(self, task_id):
        return self.executor.kill(task_id)

    def reconcile(self, task_config):
        self.executor.reconcile(task_config)

    def run(self, task_config):
        self.executor.run(task_config)
        event_queue = self.executor.get_event_queue()

        while True:
            event = event_queue.get()

            if event.kind == 'control' and \
               event.message == 'stop':
                log.info('Stop event received: {}'.format(event))
                return event

            if event.task_id != task_config.task_id:
                event_queue.put(event)
                time.sleep(1)  # hope somebody else picks it up?
                continue

            if event.terminal:
                return event

    def stop(self):
        self.executor.stop()



import decimal

import boto3.session as bsession
from boto3.dynamodb.conditions import Key
from pyrsistent import thaw

from task_processing.interfaces.persistence import Persister


class DynamoDBPersister(Persister):
    def __init__(self, table_name, endpoint_url=None, session=None):
        self.table_name = table_name
        if not session:
            session = bsession.Session()
        self.ddb_client = session.client(
            service_name='dynamodb',
            endpoint_url=endpoint_url,
        )
        self.table = session.resource(
            endpoint_url=endpoint_url,
            service_name='dynamodb'
        ).Table(table_name)

    def read(self, task_id, comparison_operator='EQ'):
        res = self.table.query(
            KeyConditionExpression=Key('task_id').eq(task_id)
        )
        return [self.item_to_event(item) for item in res['Items']]

    def write(self, event):
        if event.kind == 'control':
            return None
        return self.ddb_client.put_item(
            TableName=self.table_name,
            Item=self._event_to_item(event)['M']
        )

    def _event_to_item(self, e):
        raw = thaw(e)
        if type(raw) is dict:
            resp = {}
            for k, v in raw.items():
                if type(v) is str:
                    resp[k] = {
                        'S': v
                    }
                elif type(v) is bool:
                    resp[k] = {
                        'BOOL': v
                    }
                elif isinstance(v, (int, float)):
                    resp[k] = {
                        'N': str(v)
                    }
                elif type(v) is dict:
                    resp[k] = self._event_to_item(v)
                elif type(v) is list:
                    if len(v) > 0:
                        vals = []
                        for i in v:
                            vals.append(self._event_to_item(i))
                        resp[k] = {
                            'L': vals
                        }
            return {'M': resp}
        elif type(raw) is str:
            return {
                'S': raw
            }
        elif type(raw) in [int, float]:
            return {
                'N': str(raw)
            }
        else:
            print("Missed converting key %s type %s" % (raw, type(raw)))

    def item_to_event(self, obj):
        return self._replace_decimals(obj)

    def _replace_decimals(self, obj):
        if isinstance(obj, list):
            return [self._replace_decimals(x) for x in obj]
        elif isinstance(obj, dict):
            return {k: self._replace_decimals(v) for k, v in obj.items()}
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        else:
            return obj



from enum import Enum
from typing import Any
from typing import List

import yaml


class AutoEnum(Enum):
    """
    Enums in Python require that some methods be defined pre-construction if overrides are
    desired, so if we want to create an enum that sets the value of all members to the name
    of said members, we need to create an Enum class that overrides _generate_next_value_ such
    that when EnumBase does its magic, it sees this method.

    Based on the snippet in https://docs.python.org/3/library/enum.html#using-automatic-values
    """
    @staticmethod
    def _generate_next_value_(
        name: str,
        start: int,
        count: int,
        last_values: List[Any],
    ) -> str:
        """Override auto() to set all enum values to the name of the member"""
        return name


def get_cluster_master_by_proxy(
    proxy_prefix,
    cluster,
    services_file='/nail/etc/services/services.yaml',
):
    with open(services_file, 'r') as istream:
        services = yaml.load(istream)

    proxy_name = 'task_processing.' + proxy_prefix + '_' + cluster
    if proxy_name not in services:
        raise KeyError('Cluster master {} was not found'.format(proxy_name))
    else:
        addr_info = services[proxy_name]
        return addr_info['host'] + ':' + str(addr_info['port'])



import hashlib
import logging
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

from kubernetes.client import V1Capabilities
from kubernetes.client import V1EnvVar
from kubernetes.client import V1EnvVarSource
from kubernetes.client import V1HostPathVolumeSource
from kubernetes.client import V1NodeAffinity
from kubernetes.client import V1NodeSelector
from kubernetes.client import V1NodeSelectorRequirement
from kubernetes.client import V1NodeSelectorTerm
from kubernetes.client import V1SecretKeySelector
from kubernetes.client import V1SecurityContext
from kubernetes.client import V1Volume
from kubernetes.client import V1VolumeMount
from pyrsistent.typing import PMap
from pyrsistent.typing import PVector

from task_processing.plugins.kubernetes.types import NodeAffinityOperator
if TYPE_CHECKING:
    from task_processing.plugins.kubernetes.types import DockerVolume
    from task_processing.plugins.kubernetes.types import NodeAffinity
    from task_processing.plugins.kubernetes.types import SecretEnvSource

logger = logging.getLogger(__name__)


def get_security_context_for_capabilities(
    cap_add: PVector[str],
    cap_drop: PVector[str],
) -> Optional[V1SecurityContext]:
    """
    Helper to take lists of capabilties to add/drop and turn them into the
    corresponding Kubernetes representation.
    """
    caps = {
        capability_type: capabilities
        for (capability_type, capabilities)
        in [("add", list(cap_add)), ("drop", list(cap_drop))]
        if capabilities
    }
    if caps:
        return V1SecurityContext(capabilities=V1Capabilities(**caps))

    logger.info(
        "No capabilities found, not creating a security context"
    )
    return None


def get_kubernetes_env_vars(
    environment: PMap[str, str], secret_environment: PMap[str, 'SecretEnvSource'],
) -> List[V1EnvVar]:
    """
    Given a dict of environment variables, transform them into the corresponding Kubernetes
    representation. This function will replace any secret placeholders with the value of
    the actual secret.
    """
    env_vars = [
        V1EnvVar(name=key, value=value) for key, value
        in environment.items() if key not in secret_environment.keys()
    ]

    secret_env_vars = [
        V1EnvVar(name=key, value_from=V1EnvVarSource(
            secret_key_ref=V1SecretKeySelector(
                name=value["secret_name"],
                key=value["key"],
                optional=False,
            ),
        ),
        )
        for key, value
        in secret_environment.items()
    ]

    return env_vars + secret_env_vars


def get_sanitised_kubernetes_name(
    name: str,
    replace_dots: bool = False,
    replace_forward_slash: bool = False,
    length_limit: int = 0,
) -> str:
    """
    Helper to ensure that any names given to Kubernetes objects follow our conventions

    replace_dots is an optional parameter for objects such as Containers that cannot contain `.`s in
    their names (in contrast to objects such as Pods that can)

    replace_forward_slash is an optional parameter for objects that may contain / in their "pretty"
    names, but that cannot contain / in their Kubernetes name

    NOTE: For names exceeding the length limit, we'll truncate the name and replace the
    truncated portion with a unique suffix.
    """
    name = name.replace("_", "--")
    if name.startswith("--"):
        name = name.replace("--", "underscore-", 1)
    if replace_dots:
        name = name.replace(".", "dot-")
    if replace_forward_slash:
        name = name.replace("/", "slash-")

    if length_limit and len(name) > length_limit:
        # for names that exceed the length limit, we'll remove 6 characters from
        # the part of the name that does fit in the limit in order to replace them with
        # -- (2 characters) and then a 4 character hash (which should help ensure uniqueness
        # after truncation).
        name = (
            name[0:length_limit - 6]
            + "--"
            + hashlib.md5(name.encode("ascii")).hexdigest()[:4]
        )
    return name.lower()


def get_sanitised_volume_name(volume_name: str, length_limit: int = 0) -> str:
    """
    Helper to ensure that volume names follow our conventions (and respect an optional length limit)

    NOTE: For volume names exceeding the length limit, we'll truncate the name and replace the
    truncated portion with a unique suffix.
    """
    volume_name = volume_name.rstrip("/")
    return get_sanitised_kubernetes_name(
        volume_name,
        replace_dots=True,
        replace_forward_slash=True,
        length_limit=length_limit,
    )


def get_kubernetes_volume_mounts(volumes: PVector['DockerVolume']) -> List[V1VolumeMount]:
    """
    Given a list of volume mounts, return a list corresponding to the Kubernetes objects
    representing these mounts.
    """
    return [
        V1VolumeMount(
            mount_path=volume["container_path"],
            name=get_sanitised_volume_name(
                f"host--{volume['host_path']}",
                length_limit=63
            ),
            read_only=volume.get("mode", "RO") == "RO",
        )
        for volume in volumes
    ]


def get_pod_volumes(volumes: PVector['DockerVolume']) -> List[V1Volume]:
    """
    Given a list of volume mounts, return a list corresponding to the Kubernetes objects needed to
    tie the mounts to a Pod.
    """
    unique_volumes: Dict[str, 'DockerVolume'] = {
        get_sanitised_volume_name(f"host--{volume['host_path']}", length_limit=63): volume
        for volume in volumes
    }

    return [
        V1Volume(
            host_path=V1HostPathVolumeSource(path=volume["host_path"]),
            name=name,
        )
        for name, volume in unique_volumes.items()
    ]


def get_node_affinity(affinities: PVector["NodeAffinity"]) -> Optional[V1NodeAffinity]:
    # convert NodeAffinity into V1NodeSelectorRequirement
    match_expressions = []
    for aff in affinities:
        op = aff["operator"]
        val = aff["value"]

        # operator and value are assumed to be validated
        if op in {NodeAffinityOperator.IN, NodeAffinityOperator.NOT_IN}:
            val = [str(v) for v in val]
        elif op in {NodeAffinityOperator.GT, NodeAffinityOperator.LT}:
            val = [str(val)]
        elif op in {NodeAffinityOperator.EXISTS, NodeAffinityOperator.DOES_NOT_EXIST}:
            val = []
        else:
            continue
        match_expressions.append(
            V1NodeSelectorRequirement(key=str(aff["key"]), operator=op, values=val)
        )

    # package into V1NodeAffinity
    if not match_expressions:
        return None
    return V1NodeAffinity(
        # this means that the selectors are only used during scheduling.
        # changing it while the pod is running will not cause an eviction.
        required_during_scheduling_ignored_during_execution=V1NodeSelector(
            node_selector_terms=[V1NodeSelectorTerm(match_expressions=match_expressions)],
        ),
    )



import re
import secrets
import string
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TYPE_CHECKING

from pyrsistent import field
from pyrsistent import m
from pyrsistent import PMap
from pyrsistent import pmap
from pyrsistent import PVector
from pyrsistent import pvector
from pyrsistent import v

from task_processing.plugins.kubernetes.types import NodeAffinity
from task_processing.plugins.kubernetes.types import NodeAffinityOperator
from task_processing.plugins.kubernetes.utils import get_sanitised_kubernetes_name
if TYPE_CHECKING:
    from task_processing.plugins.kubernetes.types import DockerVolume
    from task_processing.plugins.kubernetes.types import SecretEnvSource

from task_processing.interfaces.task_executor import DefaultTaskConfigInterface

POD_SUFFIX_ALPHABET = string.ascii_lowercase + string.digits
POD_SUFFIX_LENGTH = 6
# The max length is actually 253, but https://github.com/kubernetes/kubernetes/issues/91410 means
# that the effective limit is actually:
# 255 - 63 - 37 - 2 = 153
# or (max filename length) - (max namespace length) -(pod UID length) - (separators)
# but let's give ourselves a little buffer so we'll round down a bit
MAX_POD_NAME_LENGTH = 150
VALID_POD_NAME_REGEX = '[a-z0-9]([.-a-z0-9]*[a-z0-9])?'
VALID_VOLUME_KEYS = {'mode', 'container_path', 'host_path'}
VALID_SECRET_ENV_KEYS = {'secret_name', 'key'}
VALID_CAPABILITIES = {
    "AUDIT_CONTROL",
    "AUDIT_READ",
    "AUDIT_WRITE",
    "BLOCK_SUSPEND",
    "CHOWN",
    "DAC_OVERRIDE",
    "DAC_READ_SEARCH",
    "FOWNER",
    "FSETID",
    "IPC_LOCK",
    "IPC_OWNER",
    "KILL",
    "LEASE",
    "LINUX_IMMUTABLE",
    "MAC_ADMIN",
    "MAC_OVERRIDE",
    "MKNOD",
    "NET_ADMIN",
    "NET_BIND_SERVICE",
    "NET_BROADCAST",
    "NET_RAW",
    "SETFCAP",
    "SETGID",
    "SETPCAP",
    "SETUID",
    "SYSLOG",
    "SYS_ADMIN",
    "SYS_BOOT",
    "SYS_CHROOT",
    "SYS_MODULE",
    "SYS_NICE",
    "SYS_PACCT",
    "SYS_PTRACE",
    "SYS_RAWIO",
    "SYS_RESOURCE",
    "SYS_TIME",
    "SYS_TTY_CONFIG",
    "WAKE_ALARM",
}
DEFAULT_CAPS_DROP = {
    "AUDIT_WRITE",
    "CHOWN",
    "DAC_OVERRIDE",
    "FOWNER",
    "FSETID",
    "KILL",
    "MKNOD",
    "NET_BIND_SERVICE",
    "NET_RAW",
    "SETFCAP",
    "SETGID",
    "SETPCAP",
    "SETUID",
    "SYS_CHROOT",
}
VALID_DOCKER_VOLUME_MODES = {"RW", "RO"}
REQUIRED_NODE_AFFINITY_KEYS = set(NodeAffinity.__annotations__.keys())


def _generate_pod_suffix() -> str:
    return ''.join(secrets.choice(POD_SUFFIX_ALPHABET) for i in range(POD_SUFFIX_LENGTH))


def _valid_volumes(volumes: Sequence["DockerVolume"]) -> Tuple[bool, Optional[str]]:
    for volume in volumes:
        if set(volume.keys()) != VALID_VOLUME_KEYS:
            return (
                False,
                f'Invalid volume format, must only contain following keys: '
                f'{VALID_VOLUME_KEYS}, got: {volume.keys()}'
            )
        if volume["mode"] not in VALID_DOCKER_VOLUME_MODES:
            return (
                False,
                f"Invalid mode for volume, must be one of {VALID_DOCKER_VOLUME_MODES}",
            )
    return (True, None)


def _valid_secret_envs(secret_envs: Mapping[str, "SecretEnvSource"]) -> Tuple[bool, Optional[str]]:
    # Note we are not validating existence of secret in k8s here, leave that to creation of pod
    for key, value in secret_envs.items():
        if set(value.keys()) != VALID_SECRET_ENV_KEYS:
            return (
                False,
                f'Invalid secret environment variable {key}, must only contain following keys: '
                f'{VALID_SECRET_ENV_KEYS}, got: {value.keys()}'
            )
    return (True, None)


def _valid_capabilities(capabilities: Sequence[str]) -> Tuple[bool, Optional[str]]:
    if (set(capabilities) & VALID_CAPABILITIES) != set(capabilities):
        return (
            False,
            f"Invalid capabilities - got {capabilities} but expected only values from "
            f"{VALID_CAPABILITIES}",
        )
    return (True, None)


def _valid_node_affinities(affinities: Sequence["NodeAffinity"]) -> Tuple[bool, Optional[str]]:
    for aff in affinities:
        missing_keys = REQUIRED_NODE_AFFINITY_KEYS.difference(set(aff.keys()))
        if missing_keys:
            return (
                False,
                f"Invalid node affinity: got {aff} but missing keys {missing_keys}"
            )

        op, val = aff["operator"], aff["value"]
        if op not in NodeAffinityOperator:
            valid_operators = list(o.value for o in NodeAffinityOperator)
            return (
                False,
                f"Invalid node affinity operator: got '{op}', "
                f"but expected one of: {valid_operators}",
            )

        elif (
            op in {NodeAffinityOperator.IN, NodeAffinityOperator.NOT_IN} and
            type(val) not in {list, tuple}
        ):
            return (
                False,
                "Invalid node affinity value: "
                f"got non-list value '{val}' for affinity operator '{op}'"
            )

        elif (
            op in {NodeAffinityOperator.GT, NodeAffinityOperator.LT} and
            type(val) != int
        ):
            return (
                False,
                "Invalid node affinity value: "
                f"got non-int value '{val}' for affinity operator '{op}'",
            )

    return True, None


class KubernetesTaskConfig(DefaultTaskConfigInterface):
    def __invariant__(self) -> Tuple[Tuple[bool, str], ...]:
        valid_length = len(self.pod_name) <= MAX_POD_NAME_LENGTH
        valid_name = bool(re.match(VALID_POD_NAME_REGEX, self.pod_name))

        return (
            (valid_length, f'Pod name must have up to {MAX_POD_NAME_LENGTH} characters.'),
            (valid_name, 'Must comply with Kubernetes pod naming standards.'),
        )

    uuid = field(type=str, initial=_generate_pod_suffix)  # type: ignore
    name = field(type=str, initial="default")
    # Hardcoded for the time being
    restart_policy = "Never"
    # By default, the retrying executor retries 3 times. This task option
    # overrides the executor setting.
    retries = field(
        type=int,
        factory=int,
        mandatory=False,
        invariant=lambda r: (r >= 0, 'retries >= 0')
    )

    image = field(type=str, mandatory=True)
    command = field(
        type=str,
        mandatory=True,
        invariant=lambda cmd: (cmd.strip() != '', 'empty command is not allowed')
    )
    volumes = field(
        type=PVector if not TYPE_CHECKING else PVector["DockerVolume"],
        initial=v(),
        factory=pvector,
        invariant=_valid_volumes,
    )

    cpus = field(
        type=float,
        initial=0.1,
        factory=float,
        invariant=lambda c: (c > 0, 'cpus > 0'))
    memory = field(
        type=float,
        initial=128.0,
        factory=float,
        invariant=lambda m: (m >= 32, 'mem is >= 32'))
    disk = field(
        type=float,
        initial=10.0,
        factory=float,
        invariant=lambda d: (d > 0, 'disk > 0'))
    environment = field(
        type=PMap if not TYPE_CHECKING else PMap[str, str],
        initial=m(),
        factory=pmap,
    )
    secret_environment = field(
        type=PMap if not TYPE_CHECKING else PMap[str, 'SecretEnvSource'],
        initial=m(),
        factory=pmap,
        invariant=_valid_secret_envs,
    )
    cap_add = field(
        type=PVector if not TYPE_CHECKING else PVector[str],
        initial=v(),
        factory=pvector,
        invariant=_valid_capabilities,
    )
    cap_drop = field(
        type=PVector if not TYPE_CHECKING else PVector[str],
        initial=pvector(DEFAULT_CAPS_DROP),
        factory=pvector,
        invariant=_valid_capabilities,
    )
    node_selectors = field(
        type=PMap if not TYPE_CHECKING else PMap[str, str],
        initial=m(),
        factory=pmap,
    )
    node_affinities = field(
        type=PVector if not TYPE_CHECKING else PVector["NodeAffinity"],
        initial=v(),
        factory=pvector,
        invariant=_valid_node_affinities,
    )
    labels = field(
        type=PMap if not TYPE_CHECKING else PMap[str, str],
        initial=m(),
        factory=pmap,
    )
    fs_group = field(
        type=int,
        # this is the `nobody` user at Yelp, which is what we should always be using
        # and, as such, is probably the best default to add here.
        initial=65534,
        invariant=lambda group: (0 <= group <= 65534, 'fs_group must be >= 0 and <= 65,534'),
    )

    @property
    def pod_name(self) -> str:
        return get_sanitised_kubernetes_name(
            f'{self.name}.{self.uuid}',  # type: ignore
            length_limit=MAX_POD_NAME_LENGTH,
        )

    def set_pod_name(self, pod_name: str):
        try:
            name, uuid = pod_name.rsplit('.', maxsplit=1)
        except ValueError:
            raise ValueError(f'Invalid format for pod_name {pod_name}')

        return self.set(name=name, uuid=uuid)



import uuid
from typing import Sequence
from typing import TYPE_CHECKING

from pyrsistent import field
from pyrsistent import m
from pyrsistent import PMap
from pyrsistent import pmap
from pyrsistent import PVector
from pyrsistent import pvector
from pyrsistent import v

from task_processing.interfaces.task_executor import DefaultTaskConfigInterface
from task_processing.plugins.mesos.constraints import Constraint
from task_processing.plugins.mesos.constraints import \
    valid_constraint_operator_name

VOLUME_KEYS = set(['mode', 'container_path', 'host_path'])


def valid_volumes(volumes):
    for vol in volumes:
        if set(vol.keys()) != VOLUME_KEYS:
            return (
                False,
                'Invalid volume format, must only contain following keys: '
                '{}, was: {}'.format(VOLUME_KEYS, vol.keys())
            )
    return (True, None)


def _valid_constraints(constraints):
    invalid_operators = []
    for constraint in constraints:
        operator_name = constraint.operator
        if not valid_constraint_operator_name(operator_name):
            invalid_operators.append(operator_name)
    if len(invalid_operators) > 0:
        return (
            False,
            'Invalid constraint operator names: {}'.format(operator_name)
        )
    else:
        return (True, None)


class MesosTaskConfig(DefaultTaskConfigInterface):
    def __invariant__(conf):
        return (
            (
                'image' in conf if conf.containerizer == 'DOCKER' else True,
                'Image required for chosen containerizer',
            ),
            (
                len(conf.task_id) <= 255,
                'task_id is longer than 255 chars: {}'.format(conf.task_id)
            ),
        )

    uuid = field(type=(str, uuid.UUID), initial=uuid.uuid4)  # type: ignore
    name = field(type=str, initial="default")
    # image is optional for the mesos containerizer
    image = field(type=str)
    cmd = field(type=str,
                mandatory=True,
                invariant=lambda cmd: (cmd.strip() != '', 'empty cmd'))
    cpus = field(type=float,
                 initial=0.1,
                 factory=float,
                 invariant=lambda c: (c > 0, 'cpus > 0'))
    mem = field(type=float,
                initial=32.0,
                factory=float,
                invariant=lambda m: (m >= 32, 'mem is >= 32'))
    disk = field(type=float,
                 initial=10.0,
                 factory=float,
                 invariant=lambda d: (d > 0, 'disk > 0'))
    gpus = field(type=int,
                 initial=0,
                 factory=int,
                 invariant=lambda g: (g >= 0, 'gpus >= 0'))
    timeout = field(type=float,
                    factory=float,
                    mandatory=False,
                    invariant=lambda t: (t > 0, 'timeout > 0'))
    # By default, the retrying executor retries 3 times. This task option
    # overrides the executor setting.
    retries = field(type=int,
                    factory=int,
                    mandatory=False,
                    invariant=lambda r: (r >= 0, 'retries >= 0'))
    volumes = field(type=PVector,
                    initial=v(),
                    factory=pvector,
                    invariant=valid_volumes)
    ports = field(type=(PVector[PMap] if TYPE_CHECKING else PVector),
                  initial=v(),
                  factory=pvector)
    cap_add = field(type=PVector, initial=v(), factory=pvector)
    ulimit = field(type=PVector, initial=v(), factory=pvector)
    uris = field(type=PVector, initial=v(), factory=pvector)
    # TODO: containerization + containerization_args ?
    docker_parameters = field(type=PVector, initial=v(), factory=pvector)
    containerizer = field(type=str,
                          initial='DOCKER',
                          invariant=lambda c:
                          (c == 'DOCKER' or c == 'MESOS',
                           'containerizer is docker or mesos'))
    environment = field(type=PMap, initial=m(), factory=pmap)
    offer_timeout = field(
        type=float,
        initial=60.0,
        factory=float,
        invariant=lambda t: (t > 0, 'timeout > 0')
    )
    constraints = field(
        type=(Sequence[Constraint] if TYPE_CHECKING else PVector),
        initial=v(),
        factory=lambda c: pvector((Constraint(attribute=v[0], operator=v[1],
                                              value=v[2]) for v in c)),
        invariant=_valid_constraints,
    )
    use_cached_image = field(type=bool, initial=True, factory=bool)

    @property
    def task_id(self):
        return "{}.{}".format(self.name, self.uuid)

    def set_task_id(self, task_id):
        try:
            name, uuid = task_id.rsplit('.', maxsplit=1)
        except ValueError:
            raise ValueError(f'Invalid format for task_id {task_id}')
        return self.set(name=name, uuid=uuid)



import re

from pyrsistent import field
from pyrsistent import PRecord


def equals_op(expected_value, actual_value):
    return expected_value == actual_value


def notequals_op(expected_value, actual_value):
    return expected_value != actual_value


def like_op(re_pattern, actual_value):
    return re.fullmatch(re_pattern, actual_value)


def unlike_op(re_pattern, actual_value):
    return not like_op(re_pattern, actual_value)


OPERATORS = {
    'EQUALS': equals_op,
    '==': equals_op,
    'NOTEQUALS': notequals_op,
    '!=': notequals_op,
    'LIKE': like_op,
    'UNLIKE': unlike_op,
}


def _attributes_match_constraint(attributes, constraint):
    actual_value = attributes.get(constraint.attribute)
    # If the dictionary doesn't contain an attribute from the constraint then
    # the constraint is satisfied.
    if actual_value is None:
        return True

    # The operator names have already been validated by the validator in
    # `MesosTaskConfig`, so it's guaranteed that it's in `OPERATORS`.
    return OPERATORS[constraint.operator](constraint.value, actual_value)


def attributes_match_constraints(attributes, constraints):
    # If constraints aren't specified then they are satisfied.
    if constraints is None:
        return True

    return all(_attributes_match_constraint(attributes, c)
               for c in constraints)


def valid_constraint_operator_name(name):
    operators_names = OPERATORS.keys()
    return (name in operators_names,
            '{operator} is not a valid operator, valid operators are '
            '{operators}.'.format(operator=name,
                                  operators=operators_names))


class Constraint(PRecord):
    attribute = field(type=str)
    operator = field(type=str, invariant=valid_constraint_operator_name)
    value = field(type=str)



import logging
import os
import traceback
from collections import namedtuple
from queue import Empty
from threading import Thread

from task_processing.interfaces.runner import Runner

EventHandler = namedtuple('EventHandler', ['predicate', 'cb'])

log = logging.getLogger(__name__)


class AsyncError(Exception):
    pass


class Async(Runner):
    # TODO: "callbacks" is inconsistent with the EventHandler terminology
    # above. This should either be event_handlers, or
    # EventHandler should be Callback
    def __init__(self, executor, callbacks=None):
        if not callbacks:
            raise AsyncError("must provide at least one callback")

        self.callbacks = callbacks
        self.executor = executor
        self.TASK_CONFIG_INTERFACE = executor.TASK_CONFIG_INTERFACE
        self.stopping = False

        self.callback_t = Thread(target=self.callback_loop)
        self.callback_t.daemon = True
        self.callback_t.start()

    def run(self, task_config):
        return self.executor.run(task_config)

    def kill(self, task_id):
        return self.executor.kill(task_id)

    def reconcile(self, task_config):
        self.executor.reconcile(task_config)

    def callback_loop(self):
        event_queue = self.executor.get_event_queue()

        while True:
            if self.stopping:
                return

            try:
                event = event_queue.get(True, 10)

                # TODO: have a default callback? raise exception when this
                # event is ignored?
                if event.kind == 'control' and \
                   event.message == 'stop':
                    self.stopping = True
                    continue

                for cb in self.callbacks:
                    if cb.predicate(event):
                        try:
                            cb.cb(event)
                        except Exception:
                            log.error(traceback.format_exc())
                            os._exit(1)
            except Empty:
                pass

    def stop(self):
        self.executor.stop()
        self.stopping = True
        self.callback_t.join()



import enum
from typing import Any
from typing import Dict

from kubernetes.client import V1Pod
from typing_extensions import TypedDict


class DockerVolume(TypedDict):
    host_path: str
    container_path: str
    mode: str  # XXX: Literal["RO", "RW"] once we drop older Python support


class SecretEnvSource(TypedDict):
    secret_name: str  # full name of k8s secret resource
    key: str


class EnumSet(enum.EnumMeta):
    def __contains__(cls, v):
        try:
            cls(v)
        except ValueError:
            return False
        else:
            return True


class NodeAffinityOperator(str, enum.Enum, metaclass=EnumSet):
    IN = "In"
    NOT_IN = "NotIn"
    EXISTS = "Exists"
    DOES_NOT_EXIST = "DoesNotExist"
    GT = "Gt"
    LT = "Lt"


# the value depends on operator:
# - In/NotIn requires a list
# - Exists/DoesNotExist does not expect a value
# - Gt/Lt requires an int
# the value is converted into a list of strings, which is expected by
# V1NodeSelectorRequirement.
class NodeAffinity(TypedDict):
    key: str
    operator: str
    value: Any


class PodEvent(TypedDict):
    # there are only 3 possible types for Pod events: ADDED, DELETED, MODIFIED
    # XXX: this should be typed as Literal["ADDED", "DELETED", "MODIFIED"] once we drop support
    # for older Python versions
    type: str
    object: V1Pod
    # this is just the dict-ified version of object - but it's too big to type here
    raw_object: Dict[str, Any]



from typing import Tuple
from typing import TYPE_CHECKING

import addict
from pyrsistent import field
from pyrsistent import m
from pyrsistent import PMap
from pyrsistent import pmap
from pyrsistent import PRecord
from pyrsistent import PVector
from pyrsistent import pvector
from pyrsistent import v

from task_processing.plugins.mesos.task_config import MesosTaskConfig

NUMERIC_RESOURCE = field(
    type=float,
    initial=0.0,
    factory=float,
    invariant=lambda x: (x >= 0, 'resource < 0'),
)
_NUMERIC_RESOURCES = frozenset(['cpus', 'mem', 'disk', 'gpus'])


class ResourceSet(PRecord):
    cpus = NUMERIC_RESOURCE
    mem = NUMERIC_RESOURCE
    disk = NUMERIC_RESOURCE
    gpus = NUMERIC_RESOURCE
    ports = field(type=(PVector[PMap] if TYPE_CHECKING else PVector),
                  initial=v(),
                  factory=pvector)


def get_offer_resources(offer: addict.Dict, role: str) -> ResourceSet:
    """ Get the resources from a Mesos offer

    :param offer: the payload from a Mesos resourceOffer call
    :param role: the Mesos role we want to get resources for
    :returns: a mapping from resource name -> available resources for the offer
    """
    res = ResourceSet()
    for resource in offer.resources:
        if resource.role != role:
            continue

        if resource.name in _NUMERIC_RESOURCES:
            res = res.set(resource.name, resource.scalar.value)
        elif resource.name == 'ports':
            res = res.set('ports', [pmap(r) for r in resource.ranges.range])
    return res


def allocate_task_resources(
    task_config: MesosTaskConfig,
    offer_resources: ResourceSet,
) -> Tuple[MesosTaskConfig, ResourceSet]:
    """ Allocate a task's resources to a Mesos offer

    :param task: the specification for the task to allocate
    :param offer_resources: a mapping of resource name -> available resources
        (should come from :func:`get_offer_resources`)
    :returns: a pair of (`prepared_task_config`, `remaining_resources`), where
        `prepared_task_config` is the task_config object modified with the
        actual resources consumed
    """
    for res, val in offer_resources.items():
        if res not in _NUMERIC_RESOURCES:
            continue
        offer_resources = offer_resources.set(res, val - task_config[res])

    port = offer_resources.ports[0]['begin']
    if offer_resources.ports[0]['begin'] == offer_resources.ports[0]['end']:
        avail_ports = offer_resources.ports[1:]
    else:
        new_port_range = offer_resources.ports[0].set('begin', port + 1)
        avail_ports = offer_resources.ports.set(0, new_port_range)
    offer_resources = offer_resources.set('ports', avail_ports)
    task_config = task_config.set('ports', v(m(begin=port, end=port)))
    return task_config, offer_resources


def task_fits(task: MesosTaskConfig, offer_resources: ResourceSet) -> bool:
    """ Check to see if a task fits a given offer's resources

    :param task: the task specification to check
    :param offer_resources: a mapping of resource name -> available resources
        (should come from :func:`get_offer_resources`)
    :returns: True if the offer has enough resources for the task, False otherwise
    """
    for rname, value in offer_resources.items():
        if rname in _NUMERIC_RESOURCES and task[rname] > value:
            return False
        elif rname == 'ports' and len(value) == 0:  # TODO validate port ranges
            return False

    return True



import abc
import uuid

from pyrsistent import field
from pyrsistent import PRecord


class DefaultTaskConfigInterface(PRecord):
    uuid = field(type=uuid.UUID, initial=uuid.uuid4)
    name = field(type=str, initial='default')


class TaskExecutor(metaclass=abc.ABCMeta):
    """The core interface for Task Processing
    This is the class you want to implement to add a new TaskExecutor
    """

    TASK_CONFIG_INTERFACE = DefaultTaskConfigInterface
    """
    The interface, specified as a PRecord of
    objects that you will be passing as task_configs to run
    """

    @abc.abstractmethod
    def run(self, task_config):
        """Run the supplied task

        :param task_config: An object satistfying the TASK_CONFIG_INTERFACE
        The executor should start running the provided task and return the
        task id.

        :returns str task_id: Callers get the id of the task that was run
        to check status or kill it later
        """
        pass

    @abc.abstractmethod
    def reconcile(self, task_config):
        """Request a status event for a task

        :param task_config: The task that you want to check
        """
        pass

    @abc.abstractmethod
    def kill(self, task_id):
        """Kill the specified task

        :param str task_id: The task that you want to kill
        """
        pass

    @abc.abstractmethod
    def stop(self):
        """Stop the executor stack
        """
        pass

    @abc.abstractmethod
    def get_event_queue(self):
        """Get queue of events

        :returns: TBD
        """
        pass



import logging
import time
from queue import Queue
from threading import Lock
from threading import Thread

from pyrsistent import m

from task_processing.interfaces.task_executor import TaskExecutor

log = logging.getLogger(__name__)


class RetryingExecutor(TaskExecutor):
    def __init__(self,
                 downstream_executor,
                 retry_pred=lambda e: not e.success,
                 retries=3):
        self.executor = downstream_executor
        self.retries = retries
        self.retry_pred = retry_pred

        self.task_retries = m()
        self.task_retries_lock = Lock()

        self.src_queue = downstream_executor.get_event_queue()
        self.dest_queue = Queue()
        self.stopping = False

        self.retry_thread = Thread(target=self.retry_loop)
        self.retry_thread.daemon = True
        self.retry_thread.start()

    def event_with_retries(self, event):
        return event.transform(
            ('extensions', 'RetryingExecutor/tries'),
            "{}/{}".format(
                self.task_retries[event.task_id],
                self.retries
            )
        )

    def retry(self, event):
        retries_remaining = self.task_retries[event.task_id]
        if retries_remaining <= 0:
            return False

        total_retries = self._task_or_executor_retries(event.task_config)
        log.info(
            'Retrying task {}, {} of {}, fail event: {}'.format(
                event.task_config.name, total_retries - retries_remaining + 1,
                total_retries, event.raw
            )
        )

        with self.task_retries_lock:
            self.task_retries = self.task_retries.set(
                event.task_id,
                retries_remaining - 1
            )
        self.run(event.task_config)

        return True

    def retry_loop(self):
        while True:
            while not self.src_queue.empty():
                e = self.src_queue.get()

                if e.kind != 'task':
                    self.dest_queue.put(e)
                    continue

                # This is to remove trailing '-retry*'
                original_task_id = '-'.join([item for item in
                                             e.task_id.split('-')[:-1]])

                # Check if the update is for current attempt. Discard if
                # it is not.
                if not self._is_current_attempt(e, original_task_id):
                    continue

                # Set the task id back to original task_id
                e = self._restore_task_id(e, original_task_id)

                e = self.event_with_retries(e)

                if e.terminal:
                    if self.retry_pred(e):
                        if self.retry(e):
                            continue

                    with self.task_retries_lock:
                        self.task_retries = \
                            self.task_retries.remove(e.task_id)

                self.dest_queue.put(e)

            if self.stopping:
                return

            time.sleep(1)

    def run(self, task_config):
        if task_config.task_id not in self.task_retries:
            with self.task_retries_lock:
                self.task_retries = self.task_retries.set(
                    task_config.task_id,
                    self._task_or_executor_retries(task_config)
                )
        self.executor.run(self._task_config_with_retry(task_config))

    def reconcile(self, task_config):
        self.executor.reconcile(task_config)

    def kill(self, task_id):
        # retries = -1 so that manually killed tasks can be distinguished
        with self.task_retries_lock:
            self.task_retries = self.task_retries.set(
                task_id,
                -1
            )
        return self.executor.kill(task_id)

    def stop(self):
        self.executor.stop()
        self.stopping = True
        self.retry_thread.join()

    def get_event_queue(self):
        return self.dest_queue

    def _task_config_with_retry(self, task_config):
        return task_config.set(uuid='{id}-retry{attempt}'.format(
            id=task_config.uuid,
            attempt=self.task_retries[task_config.task_id]
        ))

    def _restore_task_id(self, e, original_task_id):
        task_config = e.task_config.set(uuid='-'.join(
            [item for item in str(e.task_config.uuid).split('-')[:-1]]
        ))

        # Set the task id back to original task_id
        return e.set(
            task_id=original_task_id,
            task_config=task_config,
        )

    def _is_current_attempt(self, e, original_task_id):
        retry_suffix = '-'.join([item for item in
                                 e.task_id.split('-')[-1:]])

        # This is to extract retry attempt from retry_suffix
        # eg: if retry_suffix= 'retry2', then attempt==2
        attempt = int(retry_suffix[5:])

        # This is to reregister a task with the retry executor in the event
        # of reconciliation and attempts were lost
        with self.task_retries_lock:
            if original_task_id not in self.task_retries:
                self.task_retries = self.task_retries.set(
                    original_task_id,
                    attempt,
                )
                return True

        if attempt == self.task_retries[original_task_id]:
            return True

        return False

    def _task_or_executor_retries(self, task_config):
        return task_config.retries \
            if 'retries' in task_config else self.retries



from enum import auto
from enum import unique
from typing import Tuple

from pyrsistent import field
from pyrsistent import PRecord
from pyrsistent import PVector
from pyrsistent import pvector
from pyrsistent.typing import PVector as PVectorType

from task_processing.plugins.kubernetes.task_config import KubernetesTaskConfig
from task_processing.utils import AutoEnum


@unique
class KubernetesTaskState(AutoEnum):
    TASK_PENDING = auto()
    TASK_RUNNING = auto()
    TASK_FINISHED = auto()
    TASK_FAILED = auto()
    TASK_UNKNOWN = auto()
    TASK_LOST = auto()


class KubernetesTaskMetadata(PRecord):
    # what box this task/Pod was scheduled onto
    node_name: str = field(type=str, initial='')

    # the config used to launch this task/Pod
    task_config: KubernetesTaskConfig = field(type=KubernetesTaskConfig, mandatory=True)

    task_state: KubernetesTaskState = field(type=KubernetesTaskState, mandatory=True)
    # List of state to when that state was entered (stored as a timestamp)
    task_state_history: PVectorType[Tuple[KubernetesTaskState, float]] = field(
        type=PVector, factory=pvector, mandatory=True)



import logging
import socket
import threading
import time
from collections import defaultdict
from queue import Queue
from typing import Optional  # noqa, flake8 issue
from typing import TYPE_CHECKING

from addict import Dict
from pymesos.interface import Scheduler
from pyrsistent import field
from pyrsistent import m
from pyrsistent import PMap
from pyrsistent import pmap
from pyrsistent import PRecord
from pyrsistent import PVector
from pyrsistent import v

from task_processing.interfaces.event import control_event
from task_processing.interfaces.event import task_event
from task_processing.metrics import create_counter
from task_processing.metrics import create_timer
from task_processing.metrics import get_metric
from task_processing.plugins.mesos import metrics
from task_processing.plugins.mesos.resource_helpers import get_offer_resources


if TYPE_CHECKING:
    from .mesos_executor import MesosExecutorCallbacks  # noqa


log = logging.getLogger(__name__)


class TaskMetadata(PRecord):
    agent_id = field(type=str, initial='')
    task_config = field(type=PRecord, mandatory=True)
    task_state = field(type=str, mandatory=True)
    task_state_history = field(type=PMap, factory=pmap, mandatory=True)


class ExecutionFramework(Scheduler):
    callbacks: 'MesosExecutorCallbacks'

    def __init__(
        self,
        name,
        role,
        callbacks: 'MesosExecutorCallbacks',
        task_staging_timeout_s,
        pool=None,
        slave_blacklist_timeout_s=900,
        offer_backoff=10,
        suppress_delay=10,
        initial_decline_delay=1,
        task_reconciliation_delay=300,
        framework_id=None,
        failover_timeout=604800,  # 1 week
    ) -> None:
        self.name = name
        # wait this long for a task to launch.
        self.task_staging_timeout_s = task_staging_timeout_s
        self.pool = pool
        self.role = role
        self.callbacks = callbacks
        self.slave_blacklist_timeout_s = slave_blacklist_timeout_s
        self.offer_backoff = offer_backoff

        # TODO: why does this need to be root, can it be "mesos plz figure out"
        self.framework_info = Dict(
            user='root',
            name=self.name,
            checkpoint=True,
            role=self.role,
            failover_timeout=failover_timeout,
        )
        if framework_id:
            self.framework_info['id'] = {'value': framework_id}

        self.task_queue: Queue = Queue()
        self.event_queue: Queue = Queue()
        self._driver: Optional[Scheduler] = None
        self.are_offers_suppressed = False
        self.suppress_after = int(time.time()) + suppress_delay
        self.decline_after = time.time() + initial_decline_delay
        self._task_reconciliation_delay = task_reconciliation_delay
        self._reconcile_tasks_at = time.time() + \
            self._task_reconciliation_delay

        self.offer_decline_filter = Dict(refuse_seconds=self.offer_backoff)
        self._lock = threading.RLock()
        self.blacklisted_slaves: PVector = v()
        self.task_metadata: PMap = m()

        self._initialize_metrics()
        self._last_offer_time: Optional[float] = None
        self._terminal_task_counts = {
            'TASK_FINISHED': metrics.TASK_FINISHED_COUNT,
            'TASK_LOST': metrics.TASK_LOST_COUNT,
            'TASK_KILLED': metrics.TASK_KILLED_COUNT,
            'TASK_FAILED': metrics.TASK_FAILED_COUNT,
            'TASK_ERROR': metrics.TASK_ERROR_COUNT,
            'TASK_OFFER_TIMEOUT': metrics.TASK_OFFER_TIMEOUT,
        }

        self.driver_error = object()

        self.stopping = False
        task_kill_thread = threading.Thread(
            target=self._background_check, args=())
        task_kill_thread.daemon = True
        task_kill_thread.start()

    def call_driver(self, method, *args, **kwargs):
        if not self._driver:
            log.error(f'{method} failed: No driver')
            return self.driver_error

        try:
            return getattr(self._driver, method)(*args, **kwargs)
        except (socket.timeout, Exception) as e:
            log.warning(f'{method} failed: {str(e)}')
            return self.driver_error

    def _background_check_task(self, time_now, tasks_to_reconcile, task_id, md):
        if md.task_state != 'TASK_INITED':
            tasks_to_reconcile.append(task_id)

        if md.task_state == 'TASK_INITED':
            # give up if the task hasn't launched after
            # offer_timeout
            inited_at = md.task_state_history['TASK_INITED']
            offer_timeout = md.task_config.offer_timeout
            expires_at = inited_at + offer_timeout
            if time_now >= expires_at:
                log.warning(
                    f'Task {task_id} has been waiting for offers '
                    'for longer than configured timeout '
                    f'{offer_timeout}. Giving up and removing the '
                    'task from the task queue.'
                )
                # killing the task will also remove them from the queue
                self.kill_task(task_id)
                # we are not expecting mesos to send terminal update
                # for this task, so cleaning it up manually
                self.task_metadata = self.task_metadata.discard(
                    task_id
                )
                self.event_queue.put(
                    task_event(
                        task_id=task_id,
                        terminal=True,
                        timestamp=time_now,
                        success=False,
                        message='stop',
                        task_config=md.task_config,
                        raw='Failed due to offer timeout',
                    )
                )
                get_metric(metrics.TASK_OFFER_TIMEOUT).count(1)

        # Task is not eligible for killing or reenqueuing
        in_current_state_since = md.task_state_history[md.task_state]
        if time_now < in_current_state_since + self.task_staging_timeout_s:
            return

        if md.task_state == 'UNKNOWN':
            log.warning(
                f'Re-enqueuing task {task_id} in unknown state for '
                f'longer than {self.task_staging_timeout_s}'
            )
            # Re-enqueue task
            self.enqueue_task(md.task_config)
            get_metric(
                metrics.TASK_FAILED_TO_LAUNCH_COUNT).count(1)
        elif md.task_state == 'TASK_STAGING':
            log.warning(f'Killing stuck task {task_id}')
            self.kill_task(task_id)
            self.task_metadata = self.task_metadata.set(
                task_id,
                md.set(
                    task_state='TASK_STUCK',
                    task_state_history=md.task_state_history.set(
                        'TASK_STUCK', time_now),
                )
            )
            self.blacklist_slave(
                agent_id=self.task_metadata[task_id].agent_id,
                timeout=self.slave_blacklist_timeout_s,
            )
            get_metric(metrics.TASK_STUCK_COUNT).count(1)
        elif md.task_state == 'TASK_STUCK':
            t = time.time()
            # 10s since last iteration + time we spent in current one
            time_delta = 10 + t - time_now
            # seconds since task was put in TASK_STUCK state
            time_stuck = t - md.task_state_history['TASK_STUCK']
            # seconds since `time_stuck` crossed another hour
            # boundary
            hour_rolled = time_stuck % 3600

            # if `time_stuck` crossed hour boundary since last
            # background check - lets re-send kill request
            if hour_rolled < time_delta:
                hours_stuck = time_stuck // 3600
                log.warning(
                    f'Task {task_id} is stuck, waiting for terminal '
                    f'state for {hours_stuck}h, sending another kill'
                )
                self.kill_task(task_id)

    def _background_check(self):
        while True:
            if self.stopping:
                return

            time_now = time.time()
            tasks_to_reconcile = []
            with self._lock:
                for task_id, md in self.task_metadata.items():
                    self._background_check_task(
                        time_now,
                        tasks_to_reconcile,
                        task_id,
                        md,
                    )

            self._reconcile_tasks(
                [Dict({'task_id': Dict({'value': task_id})}) for
                    task_id in tasks_to_reconcile]
            )
            elapsed = time.time() - time_now
            log.info(f'background check done in {elapsed}s')
            get_metric(metrics.BGCHECK_TIME_TIMER).record(elapsed)
            time.sleep(10)

    def reconcile_task(self, task_config):
        task_id = task_config.task_id
        with self._lock:
            if task_id in self.task_metadata:
                md = self.task_metadata[task_id]
                self.task_metadata = self.task_metadata.set(
                    task_id,
                    md.set(
                        task_state='TASK_RECONCILING',
                        task_state_history=md.task_state_history.set(
                            'TASK_RECONCILING', time.time()),
                    )
                )
            else:
                log.info(f'Adding {task_id} to metadata for reconciliation')
                self.task_metadata = self.task_metadata.set(
                    task_id,
                    TaskMetadata(
                        task_config=task_config,
                        task_state='TASK_RECONCILING',
                        task_state_history=m(TASK_RECONCILING=time.time()),
                    ),
                )
        self._reconcile_tasks([
            Dict({'task_id': Dict({'value': task_id})})
        ])

    def _reconcile_tasks(self, tasks_to_reconcile):
        if time.time() < self._reconcile_tasks_at:
            return

        log.info(f'Reconciling following tasks {tasks_to_reconcile}')

        if len(tasks_to_reconcile) > 0:
            self.call_driver('reconcileTasks', tasks_to_reconcile)

        self._reconcile_tasks_at += self._task_reconciliation_delay

    def offer_matches_pool(self, offer):
        if self.pool is None:
            # If pool is not specified, then we can accept offer from any agent
            return True, None

        for attribute in offer.attributes:
            if attribute.name == "pool":
                return attribute.text.value == self.pool, attribute.text.value

        return False, None

    def kill_task(self, task_id):
        tmp_list = []
        flag = False
        with self._lock:
            while not self.task_queue.empty():
                t = self.task_queue.get()
                if task_id == t.task_id:
                    flag = True
                    self.task_metadata = self.task_metadata.discard(task_id)
                else:
                    tmp_list.append(t)

            for t in tmp_list:
                self.task_queue.put(t)

        if flag is False:
            if self.call_driver('killTask', Dict(value=task_id)) is self.driver_error:
                return False

        return True

    def blacklist_slave(self, agent_id, timeout):
        with self._lock:
            # A new entry is appended even if the agent is being blacklisted.
            # This is equivalent to restarting the blacklist timer.
            log.info(f'Blacklisting slave: {agent_id} for {timeout} seconds.')
            self.blacklisted_slaves = self.blacklisted_slaves.append(agent_id)
            get_metric(metrics.BLACKLISTED_AGENTS_COUNT).count(1)

        unblacklist_thread = threading.Thread(
            target=self.unblacklist_slave,
            kwargs={'timeout': timeout, 'agent_id': agent_id},
        )
        unblacklist_thread.daemon = True
        unblacklist_thread.start()

    def unblacklist_slave(self, agent_id, timeout):
        time.sleep(timeout)
        log.info(
            f'Unblacklisting slave: {agent_id}'
        )
        with self._lock:
            self.blacklisted_slaves = self.blacklisted_slaves.remove(agent_id)

    def enqueue_task(self, task_config):
        with self._lock:
            # task_state and task_state_history get reset every time
            # a task is enqueued.
            self.task_metadata = self.task_metadata.set(
                task_config.task_id,
                TaskMetadata(
                    task_config=task_config,
                    task_state='TASK_INITED',
                    task_state_history=m(TASK_INITED=time.time()),
                )
            )
            # Need to lock on task_queue to prevent enqueues when getting
            # tasks to launch
            self.task_queue.put(task_config)

            if self.are_offers_suppressed:
                if self.call_driver('reviveOffers') is not self.driver_error:
                    self.are_offers_suppressed = False
                    log.info('Reviving offers because we have tasks to run.')

        get_metric(metrics.TASK_ENQUEUED_COUNT).count(1)

    def launch_tasks_for_offer(self, offer, tasks_to_launch) -> bool:
        mesos_protobuf_tasks = [
            self.callbacks.make_mesos_protobuf(
                task_config, offer.agent_id.value, self.role)
            for task_config in tasks_to_launch
            if task_config.task_id in self.task_metadata
        ]
        if not mesos_protobuf_tasks:
            return False

        launched = True
        launch_time = time.time()
        if self.call_driver('launchTasks', offer.id, mesos_protobuf_tasks) is self.driver_error:
            tasks = ', '.join(task.task_id for task in tasks_to_launch)
            log.warning(
                f'Failed to launch: {tasks}, moving them to UNKNOWN state')
            get_metric(metrics.TASK_LAUNCH_FAILED_COUNT).count(1)
            launched = False

        # 'UNKNOWN' state is for internal tracking. It will not be
        # propogated to users.
        current_task_state = 'TASK_STAGING' if launched else 'UNKNOWN'

        for task in tasks_to_launch:
            md = self.task_metadata.get(task.task_id)
            if not md:
                log.warning(
                    f'trying to launch task {task.task_id}, but it is not in task metadata.'
                    f'current keys in task_metadata: {self.task_metadata.keys()}'
                )
                continue
            self.task_metadata = self.task_metadata.set(
                task.task_id,
                md.set(
                    task_state=current_task_state,
                    task_state_history=md.task_state_history.set(
                        current_task_state, launch_time),
                    agent_id=str(offer.agent_id.value),
                )
            )

            get_metric(metrics.TASK_QUEUED_TIME_TIMER).record(
                launch_time - md.task_state_history['TASK_INITED']
            )

            # Emit the staging event for successful launches
            if launched:
                self.event_queue.put(
                    self.callbacks.handle_status_update(
                        Dict(state='TASK_STAGING', offer=offer),
                        md.task_config,
                    )
                )
                get_metric(metrics.TASK_LAUNCHED_COUNT).count(1)

        return launched

    def stop(self):
        self.stopping = True

    # TODO: add mesos cluster dimension when available
    def _initialize_metrics(self):
        default_dimensions = {
            'framework_name': '.'.join(self.name.split()[:2]),
            'framework_role': self.role
        }

        counters = [
            metrics.TASK_LAUNCHED_COUNT,                 metrics.TASK_FINISHED_COUNT,
            metrics.TASK_FAILED_COUNT,                   metrics.TASK_KILLED_COUNT,
            metrics.TASK_LOST_COUNT,                     metrics.TASK_ERROR_COUNT,
            metrics.TASK_ENQUEUED_COUNT,                 metrics.TASK_INSUFFICIENT_OFFER_COUNT,
            metrics.TASK_STUCK_COUNT,                    metrics.BLACKLISTED_AGENTS_COUNT,
            metrics.TASK_LOST_DUE_TO_INVALID_OFFER_COUNT,
            metrics.TASK_LAUNCH_FAILED_COUNT,            metrics.TASK_FAILED_TO_LAUNCH_COUNT,
            metrics.TASK_OFFER_TIMEOUT,
        ]
        for cnt in counters:
            create_counter(cnt, default_dimensions)

        timers = [
            metrics.OFFER_DELAY_TIMER, metrics.TASK_QUEUED_TIME_TIMER,
            metrics.BGCHECK_TIME_TIMER
        ]
        for tmr in timers:
            create_timer(tmr, default_dimensions)

    ####################################################################
    #                   Mesos driver hooks go here                     #
    ####################################################################
    def offerRescinded(self, driver, offerId):
        # TODO(sagarp): Executor should be able to deal with this.
        log.warning(f'Offer {offerId} rescinded')

    def error(self, driver, message):
        event = control_event(raw=message)

        # TODO: have a mapper function similar to translator of task events
        if message == 'Framework has been removed':
            event = event.set(message='stop')
        else:
            event = event.set(message='unknown')

        self.event_queue.put(event)

    def slaveLost(self, drive, slaveId):
        log.warning(f"Slave lost: {str(slaveId)}")

    def registered(self, driver, frameworkId, masterInfo):
        self._driver = driver
        event = control_event(
            raw={
                'master_info': masterInfo,
                'framework_id': frameworkId,
            },
            message='registered',
        )
        self.event_queue.put(event)
        log.info(
            f"Registered with framework ID {frameworkId.value} and role {self.role}"
        )

    def reregistered(self, driver, masterInfo):
        self._driver = driver
        log.warning(f"Re-registered to {masterInfo} with role {self.role}")

    def resourceOffers(self, driver, offers) -> None:
        self._driver = driver

        current_offer_time = time.time()
        if self._last_offer_time is not None:
            get_metric(metrics.OFFER_DELAY_TIMER).record(
                current_offer_time - self._last_offer_time
            )
        self._last_offer_time = current_offer_time

        # Give user some time to enqueue tasks
        if self.task_queue.empty() and current_offer_time < self.decline_after:
            time.sleep(self.decline_after - current_offer_time)

        declined: dict = defaultdict(list)
        declined_offer_ids = []
        accepted = []

        with self._lock:
            if self.task_queue.empty():
                # Always suppress offers when there is nothing to run
                if self.call_driver('suppressOffers') is not self.driver_error:
                    self.are_offers_suppressed = True
                    log.info("Suppressing offers, no more tasks to run.")

                for offer in offers:
                    declined['no tasks'].append(offer.id.value)
                    declined_offer_ids.append(offer.id)

                self.call_driver(
                    'declineOffer', declined_offer_ids, self.offer_decline_filter)
                log.info(
                    f"Offers declined because of no tasks: {','.join(declined['no tasks'])}")
                return

        with_maintenance_window = [
            offer for offer in offers if offer.unavailability
        ]

        for offer in with_maintenance_window:
            start_time = offer.unavailability.start['nanoseconds']
            completion_time = int(
                (start_time + offer.unavailability.duration['nanoseconds'])
                / 1000000000
            )
            now = int(time.time())
            duration = completion_time - now
            if duration > 0:
                self.blacklist_slave(
                    agent_id=offer.agent_id.value,
                    timeout=duration,
                )

        without_maintenance_window = [
            offer for offer in offers if offer not in with_maintenance_window
        ]
        for offer in without_maintenance_window:
            with self._lock:
                if offer.agent_id.value in self.blacklisted_slaves:
                    declined['blacklisted'].append(
                        f'offer {offer.id.value} agent {offer.agent_id.value}'
                    )
                    declined_offer_ids.append(offer.id)
                    continue

            offer_pool_match, offer_pool = self.offer_matches_pool(offer)
            if not offer_pool_match:
                log.info(
                    f"Declining offer {offer.id.value}, required pool "
                    f"{self.pool} doesn't match offered pool {offer_pool}"
                )
                declined['bad pool'].append(offer.id.value)
                declined_offer_ids.append(offer.id)
                continue

            # Need to lock here even though we are only reading the task_queue, since
            # we are predicating on the queue's emptiness. If not locked, other
            # threads can continue enqueueing, and we never terminate the loop.
            task_configs = []
            with self._lock:
                while not self.task_queue.empty():
                    task_configs.append(self.task_queue.get())

                offer_resources = get_offer_resources(offer, self.role)
                offer_attributes = {
                    attribute.name: attribute.text.value
                    for attribute in offer.attributes
                }
                log.info(
                    f'Received offer {offer.id.value} for role {self.role}: {offer_resources}')
                tasks_to_launch, tasks_to_defer = self.callbacks.get_tasks_for_offer(
                    task_configs,
                    offer_resources,
                    offer_attributes,
                    self.role,
                )

                for task in tasks_to_defer:
                    self.task_queue.put(task)
                get_metric(metrics.TASK_INSUFFICIENT_OFFER_COUNT).count(len(tasks_to_defer))

                if len(tasks_to_launch) == 0:
                    declined['bad resources'].append(offer.id.value)
                    declined_offer_ids.append(offer.id)
                    continue

                ignored_tasks = ','.join(
                    task_config.task_id
                    for task_config in tasks_to_launch
                    if task_config.task_id not in self.task_metadata
                )
                if ignored_tasks:
                    log.warning(
                        f'ignoring tasks not in metadata: {ignored_tasks}')

                tasks_to_launch = [
                    task_config
                    for task_config in tasks_to_launch
                    if task_config.task_id in self.task_metadata
                ]

                if len(tasks_to_launch) == 0:
                    declined['nothing to launch'].append(offer.id.value)
                    declined_offer_ids.append(offer.id)
                elif not self.launch_tasks_for_offer(offer, tasks_to_launch):
                    declined['launch failed'].append(offer.id.value)
                    declined_offer_ids.append(offer.id)
                else:
                    accepted.append(
                        f'offer: {offer.id.value} '
                        f'agent: {offer.agent_id.value} '
                        f'tasks: {len(tasks_to_launch)}'
                    )

        if len(declined_offer_ids) > 0:
            self.call_driver(
                'declineOffer', declined_offer_ids, self.offer_decline_filter
            )
        for reason, items in declined.items():
            log.info(f"Offers declined because {reason}: {', '.join(items)}")
        if accepted:
            log.info(f"Offers accepted: {', '.join(accepted)}")

    def statusUpdate(self, driver, update) -> None:
        self._driver = driver

        task_id = update.task_id.value
        task_state = str(update.state)
        log.info(f"Task update {task_state} received for task {task_id}")

        if task_id not in self.task_metadata:
            # We assume that a terminal status update has been
            # received for this task already.
            log.info('Ignoring this status update because a terminal status '
                     'update has been received for this task already.')
            self.call_driver('acknowledgeStatusUpdate', update)
            return

        md = self.task_metadata[task_id]

        # If we attempt to accept an offer that has been invalidated by
        # master for some reason such as offer has been rescinded or we
        # have exceeded offer_timeout, then we will get TASK_LOST status
        # update back from mesos master.
        if task_state == 'TASK_LOST' and str(update.reason) == \
                'REASON_INVALID_OFFERS':
            # This task has not been launched. Therefore, we are going to
            # reenqueue it. We are not propogating any event up to the
            # application.
            log.warning('Received TASK_LOST from mesos master because we '
                        'attempted to accept an invalid offer. Going to '
                        f're-enqueue this task {task_id}')
            # Re-enqueue task
            self.enqueue_task(md.task_config)
            get_metric(metrics.TASK_LOST_DUE_TO_INVALID_OFFER_COUNT).count(1)
            self.call_driver('acknowledgeStatusUpdate', update)
            return

        # Record state changes, send a new event and emit metrics only if the
        # task state has actually changed.
        if md.task_state != task_state:
            with self._lock:
                self.task_metadata = self.task_metadata.set(
                    task_id,
                    md.set(
                        task_state=task_state,
                        task_state_history=md.task_state_history.set(
                            task_state, time.time()),
                    )
                )

            self.event_queue.put(
                self.callbacks.handle_status_update(update, md.task_config),
            )

            if task_state in self._terminal_task_counts:
                with self._lock:
                    self.task_metadata = self.task_metadata.discard(task_id)
                get_metric(self._terminal_task_counts[task_state]).count(1)

        # We have to do this because we are not using implicit
        # acknowledgements.
        self.call_driver('acknowledgeStatusUpdate', update)



from task_processing.plugins.mesos.mesos_executor import MesosExecutor


class MesosPodExecutor(MesosExecutor):
    def __init__(self, role, *args, **kwargs) -> None:
        raise NotImplementedError



import logging
import threading
from typing import Callable
from typing import List
from typing import NamedTuple
from typing import Tuple

import addict
from pymesos import MesosSchedulerDriver

from task_processing.interfaces.event import Event
from task_processing.interfaces.task_executor import TaskExecutor
from task_processing.plugins.mesos.execution_framework import ExecutionFramework
from task_processing.plugins.mesos.resource_helpers import ResourceSet
from task_processing.plugins.mesos.task_config import MesosTaskConfig


class MesosExecutorCallbacks(NamedTuple):
    get_tasks_for_offer: Callable[
        [List[MesosTaskConfig], ResourceSet, dict, str],
        Tuple[List[addict.Dict], List[MesosTaskConfig]]
    ]
    handle_status_update: Callable[
        [addict.Dict, MesosTaskConfig],
        Event,
    ]
    make_mesos_protobuf: Callable[
        [MesosTaskConfig, str, str],
        addict.Dict,
    ]


class MesosExecutor(TaskExecutor):
    def __init__(
        self,
        role: str,
        callbacks: MesosExecutorCallbacks,
        pool=None,
        principal='taskproc',
        secret=None,
        mesos_address='127.0.0.1:5050',
        initial_decline_delay=1.0,
        framework_name='taskproc-default',
        framework_staging_timeout=240,
        framework_id=None,
        failover=False,
    ) -> None:
        """
        Constructs the instance of a task execution, encapsulating all state
        required to run, monitor and stop the job.

        TODO param docstrings
        """

        self.logger = logging.getLogger(__name__)
        self.role = role
        self.failover = failover

        self.execution_framework = ExecutionFramework(
            role=role,
            pool=pool,
            name=framework_name,
            callbacks=callbacks,
            task_staging_timeout_s=framework_staging_timeout,
            initial_decline_delay=initial_decline_delay,
            framework_id=framework_id,
        )

        # TODO: Get mesos master ips from smartstack
        self.driver = MesosSchedulerDriver(
            sched=self.execution_framework,
            framework=self.execution_framework.framework_info,
            use_addict=True,
            master_uri=mesos_address,
            implicit_acknowledgements=False,
            principal=principal,
            secret=secret,
            failover=failover,
        )

        # start driver thread immediately
        self.stopping = False
        self.driver_thread = threading.Thread(
            target=self._run_driver, args=())
        self.driver_thread.daemon = True
        self.driver_thread.start()

    def _run_driver(self):
        while not self.stopping:
            self.driver.run()
            self.logger.warning('Driver stopped, starting again')

    def run(self, task_config):
        self.execution_framework.enqueue_task(task_config)

    def reconcile(self, task_config):
        self.execution_framework.reconcile_task(task_config)

    def kill(self, task_id):
        return self.execution_framework.kill_task(task_id)

    def stop(self):
        self.stopping = True
        self.execution_framework.stop()
        self.driver.stop(failover=self.failover)
        self.driver.join()

    def get_event_queue(self):
        return self.execution_framework.event_queue
